/* xio.c -- safe versions of the system io primitives. It also includes a
   new version of the readlink system call that computes the number of
   characters required to hold the entire link. */

/* Copyright (C) 1993, 1994, 1995 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Written by Tudor Hulubei and Andrei Pitis.  */


#include <sys/types.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include <errno.h>

/* Not all systems declare ERRNO in errno.h... and some systems #define it! */
#if !defined (errno)
extern int errno;
#endif /* !errno */


#include "xmalloc.h"
#include "xio.h"


int xread(fd, buf, count)
    int fd;
    char *buf;
    size_t count;
{
    int chars;
    int old_errno;

    if (count <= 0)
	return count;

#ifdef EINTR
    old_errno = errno;

    do
	chars = read(fd, buf, count);
    while (chars < 0 && errno == EINTR);

    errno = old_errno;
    return chars;
#else
    return read(fd, buf, count);
#endif
}


int xwrite(fd, buf, count)
    int fd;
    char *buf;
    size_t count;
{
    int chars;
    int old_errno;

    if (count <= 0)
	return count;

#ifdef EINTR
    old_errno = errno;

    do
	chars = write(fd, buf, count);
    while (chars < 0 && errno == EINTR);

    errno = old_errno;
    return chars;
#else
    return read(fd, buf, count);
#endif
}


int __xreadlink(path, buf, size)
    const char *path;
    char *buf;
    size_t size;
{
    int chars;
    int old_errno;

    if (size <= 0)
	return size;

#ifdef EINTR
    old_errno = errno;

    do
    {
	chars = readlink(path, buf, size);
	old_errno = size;
    }
    while (chars < 0 && errno == EINTR);

    errno = old_errno;
    return chars;
#else
    return read(fd, buf, count);
#endif
}


int xreadlink(filename)
    const char *filename;
{
    int size = 100;

    for (;;)
    {
	char *buffer = xmalloc(size);
	int  nchars  = __xreadlink(filename, buffer, size);

	if (nchars < size)
	{
	    free(buffer);
	    return nchars;
	}

	free(buffer);
	size *= 2;
    }
}


int xfstat(filedes, buf)
    int filedes;
    struct stat *buf;
{
    int result;

    do
	result = fstat(filedes, buf);
    while (result < 0 && errno == EINTR);

    return result;
}


int xstat(filename, buf)
    const char *filename;
    struct stat *buf;
{
    int result;

    do
	result = stat(filename, buf);
    while (result < 0 && errno == EINTR);

    return result;
}


int xlstat(filename, buf)
    const char *filename;
    struct stat *buf;
{
    int result;

    do
#ifdef HAVE_LSTAT
	result = lstat(filename, buf);
#else
        result = stat(filename, buf);
#endif /* HAVE_LSTAT */
    while (result < 0 && errno == EINTR);

    return result;
}
